// emacs: this is -*- c++ -*-
//
//   @file    fillCache.h        
//
//                   
//  
//   Copyright (C) 2016 M.Sutton (sutt@cern.ch)    
//
//   $Id: fillCache.h, v0.0   Wed 24 Aug 2016 12:18:24 CEST sutt $


#ifndef  FILLCACHE_H
#define  FILLCACHE_H

#include <iostream>


#include "appl_grid/appl_grid.h"

#include "mapCache.h"
#include "vecCache.h"

// typedef vecCache cache_t;
// typedef mapCache cache_t;



extern "C" bool getunitphase_();
extern "C" int currentcacheid_();





template<typename T>
class fillCache {

public:

  typedef std::map<appl::grid*, T*>  cache_t;

public:

  fillCache( bool warmup=false ) : mwarmup(warmup) { } 

  virtual ~fillCache() { } 

  //  void fill( appl::grid* grid, 
  //	     int nproc, double obs, int order, int iprocess, 
  //	     double x1, double x2, double Q2, 
  //	     double fullweight );
  

  T* findcache( appl::grid* gptr, int nproc=0 ) {
    
    typename cache_t::iterator itr = mcachemap.find(gptr);
    if ( itr!=mcachemap.end() ) return itr->second;
    
    if ( nproc>0 ) { 
      //      std::pair< std::map<appl::grid*, T*>::iterator, bool > status = mcachemap.insert( std::map<appl::grid*, T*>::value_type( gptr, new T( nproc ) ) );
      std::pair< typename cache_t::iterator, bool > status = mcachemap.insert( typename cache_t::value_type( gptr, new T( nproc ) ) );
      return status.first->second;
    }
    
    return 0;
  }
  

  void fill( appl::grid* grid, 
	     int nproc, double obs, int order, int iprocess, 
	     double x1, double x2, double Q2, 
	     double fullweight ) { 
    
    cache_t* mpc = findcache( grid, nproc );
    
    /// have the kinematics changed ?
    
    bool mupdate = mpc->changed( obs, order, x1, x2, Q2 );
    
    /// std::cout << "cache " << intproc << "\t " << iprocess << "\tmupdate: " << mupdate << "\tobs " << obs 
    //            << "\tkin  " << x1 << "  " << x2 << " " << Q2 << "\twgt " << fullweight;// << std::endl;
    
    /// if they have then flush the buffer and update the grid
    
    //  std::cout << "\tiprocess " << iprocess << "\tupdate: " << mupdate << "\tobs " << obs
    //	    << "\t(" << x1 << "  " << x2 << " " << Q2 << ")\twgt " << fullweight << std::endl;
    
    if ( mupdate ) { 
      
      //    if ( dbg ) std::cout << "\tiprocess " << iprocess << "\tupdate: " << mupdate << "\tobs " << obs << std::endl;
      
      if ( mpc->filled() ){
	
	///	if ( dbg ) std::cout << "\tflushing: mpc ";
	///	if ( dbg ) std::cout << "cache size " << mpc->size() << std::endl;
	
	typename cache_t::const_iterator itr  = mpc->begin();
	typename cache_t::const_iterator iend = mpc->end();
	while( itr!=iend ) {
	  //	std::cout << "\tflushing vector size " << itr->second.size() << std::endl;
	  if (  itr->second.size()==mpc->mnproc ) { 
	    ///   if ( dbg ) std::cout << "ix" << "\tflushing obs " << itr->first << std::endl; 
	    if ( mwarmup ) grid->fill_phasespace(  mpc->mx1, mpc->mx2, mpc->mQ2, itr->first,  &itr->second[0], mpc->morder );
	    else           grid->fill_grid(         mpc->mx1, mpc->mx2, mpc->mQ2, itr->first,  &itr->second[0], mpc->morder );
	  }
	  itr++;
	}
	
      }
      
      mpc->update( obs, order, x1, x2, Q2 );
      
      mcached = 0;
      
    }
    
    ///    if ( dbg ) std::cout << std::endl;
    
    /// update the cache with the new values
    
    std::vector<double>& mwgt = mpc->weights( obs ); 
    
    mwgt.at( iprocess ) += fullweight; 
    
    mcached++;
    
    ///    If ( dbg ) std::cout << "fill: " << mpc->weights( obs ) << std::endl;
    
  }
  

private:

  bool    mwarmup;

  cache_t mcachemap;
  
  //  std::map<appl::grid*, T*>  mcachemap;

  int mcached = 0;

};


template<typename T>
inline std::ostream& operator<<( std::ostream& s, const fillCache<T>& _f ) { 
  return s;
}


#endif  // FILLCACHE_H 










